from typing import Tuple,Dict,Union,List,Optional
import math
import h5py
from sympy import sympify

#these are as defined in Nektar User Guide under Expressions
#https://doc.nektar.info/userguide/latest/user-guidese13.html
MATH_CONSTANTS: Dict[str,float] = {
    "E": math.e,
    "PI": math.pi,
    "GAMMA": 0.57721566490153286060,
    "DEG": 180 / math.pi,  # Degrees per radian
    "PHI": (1 + math.sqrt(5)) / 2,  # Golden ratio
    "LOG2E": math.log2(math.e),
    "LOG10E": math.log10(math.e),
    "LN2": math.log(2),
    "PI_2": math.pi / 2,
    "PI_4": math.pi / 4,
    "1_PI": 1 / math.pi,
    "2_PI": 2 / math.pi,
    "2_SQRTPI": 2 / math.sqrt(math.pi),
    "SQRT2": math.sqrt(2),
    "SQRT1_2": math.sqrt(0.5),
}

def get_both_sides_of_equals(text_with_equals: str) -> Tuple[str, str]:
    """Given a string "LHS = RHS", acquire LHS and RHS

    Args:
        text_with_equals (str): String containing an equality "LHS = RHS"

    Returns:
        Tuple[str, str]: Returns a tuple containing LHS string and RHS string, trimmed of leading and trailing whitespace
    """
    left, right = text_with_equals.split("=")
    return left.strip(), right.strip()

def safe_resolve(param_name: str, params: Dict[str, str], seen: set) -> Union[float, str]:
    """Recursively resolves and evaluates expressions in params. For example, the following:
    data = {"PARAM1": "10","PARAM2" : "PARAM1"}
    will be resolved into the following form due to the references:
    expected = {"PARAM1": 10,"PARAM2" : 10}
    
    Mathematical operators can also be evaluated, the following also works:
    data = {"PARAM1": "10","PARAM2" : "PARAM1 + 5"} -> expected = {"PARAM1": 10,"PARAM2" : 15}

    Args:
        param_name (str): The name of the parameter to resolve.
        params (Dict[str, str]): Dictionary of parameters mapping names to expressions.
        seen (set): Set of already visited parameters (to prevent cycles).
    
    Returns:
        Union[float, str]: Evaluated numerical result or original string if resolution fails.
    """
    if param_name in seen:
        raise ValueError(f"Circular reference detected: {' -> '.join(seen)} -> {param_name}")

    seen.add(param_name)  # Track visited parameters
    
    # Get the value (expression) of the parameter
    expr = params.get(param_name, param_name)
    
    # Try converting to number directly
    try:
        return float(expr)
    except ValueError:
        pass  # Not a direct number, so evaluate it

    # Replace references to other parameters
    for key in params:
        if key in expr:
            resolved_value = safe_resolve(key, params, seen.copy())
            expr = expr.replace(key, str(resolved_value))

    # Evaluate safely using sympy
    try:
        return float(sympify(expr, locals=MATH_CONSTANTS))
    except Exception:
        return expr  # Return as string if evaluation fails

def evaluate_parameters(params: Dict[str, str]) -> Dict[str, float]:
    """Evaluates all parameters in a dictionary. For example, the following:
    data = {"PARAM1": "10","PARAM2" : "PARAM1"}
    will be resolved into the following form due to the references:
    expected = {"PARAM1": 10,"PARAM2" : 10}
    
    Mathematical operators can also be evaluated, the following also works:
    data = {"PARAM1": "10","PARAM2" : "PARAM1 + 5"} -> expected = {"PARAM1": 10,"PARAM2" : 15}

    Args:
        params (Dict[str, str]): Dictionary mapping parameter names to expressions.
    
    Returns:
        Dict[str, float]: Evaluated parameters with numerical values.
    """
    evaluated = {}
    for param in params:
        evaluated[param] = safe_resolve(param, params, set())
    return evaluated

def get_all_files_with_extension(files: List[str],extension: str) -> List[str]:
    """Return all files in a list with the specified file extension

    Args:
        files (List[str]): List containing the files
        extension (str): File extension to be found

    Returns:
        List[str]: List of files containing the extension
    """
    # Ensure extension starts with "."
    if not extension.startswith("."):
        extension = f".{extension}"
    
    return [f for f in files if f.lower().endswith(extension.lower())]

def get_hdf5_groups_with_depth_limit(hdf5_file: h5py.File,max_depth: int,start_path: str = "",max_groups: int = 100) -> List[str]:
    """Traverses the HDF5 hierarchy and returns group paths up to a specified depth. Will find "" as first group.

    Args:
        hdf5_file (h5py.File): An open h5py.File object.
        max_depth (int): The maximum depth to traverse.
        start_path (str): the path to start the search from.
        max_groups (int): If defined, specifies maximum number of groups to be found. Once exceeded, function will return.

    Returns:
        List[str]: A list of group paths.
    """
    group_paths = []

    def _traverse(group: h5py.Group | h5py.Dataset, current_path, current_depth):
        if current_depth > max_depth:
            return

        if len(group_paths) >= max_groups:
            return
        
        group_paths.append(current_path)

        for name, obj in group.items():
            if isinstance(obj, h5py.Group):
                new_path = f"{current_path}/{name}" if current_path else name
                _traverse(obj, new_path, current_depth + 1)

    if start_path == "":
        _traverse(hdf5_file, "", 0)
    else:
        _traverse(hdf5_file[start_path], start_path, 0)

    return group_paths

def get_hdf5_datasets_with_depth_limit(hdf5_file: h5py.File,max_depth: int,start_path: str = "",max_datasets: int=100) -> List[str]:
    """Traverses the HDF5 hierarchy and returns dataset paths up to a specified depth.

    Args:
        hdf5_file (h5py.File): An open h5py.File object.
        max_depth (int): The maximum depth to traverse.
        start_path (str): the path to start the search from.
        max_groups (int): If defined, specifies maximum number of groups to be found. Once exceeded, function will return.

    Returns:
        List[str]: A list of group paths.
    """
    dataset_paths = []

    def _traverse(group, current_path, current_depth):
        if current_depth > max_depth:
            return

        for name, obj in group.items():
            if len(dataset_paths) >= max_datasets:
                return
            new_path = f"{current_path}/{name}" if current_path else name
            if isinstance(obj, h5py.Dataset):
                dataset_paths.append(new_path)
            elif isinstance(obj, h5py.Group):
                _traverse(obj, new_path, current_depth + 1)

    if start_path == "":
        _traverse(hdf5_file, "", 0)
    else:
        _traverse(hdf5_file[start_path], start_path, 0)

    return dataset_paths

if __name__ == "__main__":
    params = {
        "A": "2",
        "NEK": "A + PI_2",
        "C": "NEK*2"
    }

    print(evaluate_parameters(params))




import ttkbootstrap as ttk
import tkinter as tk
from ttkbootstrap.constants import *
import logging
import os
from NekUpload.frontend.components.settings_manager import SettingsManager
class SettingScene(ttk.Frame):
    def __init__(self,parent,settings_manager: SettingsManager):
        super().__init__(parent)
        self.settings_manager = settings_manager

        self.columnconfigure(0,weight=1)

        self.repository_info: ttk.LabelFrame = self._get_repository_info(self)
        self.repository_info.grid(row=0,column=0,columnspan=2,padx=10,pady=10,sticky=NSEW)

        self.api_key_info: ttk.LabelFrame = self._get_api_key(self)
        self.api_key_info.grid(row=1,column=0,padx=10,columnspan=2,pady=10,sticky=NSEW)

        #add a save button to save the settings configuration
        warning_label = ttk.Label(
            master=self,
            text="REMEMBER TO SAVE YOUR SETTINGS -->",
            bootstyle=DANGER,
            anchor="e"
        )
        warning_label.grid(row=2,column=0,sticky=E)


        self.save_settings_button = ttk.Button(
            master=self,
            text="SAVE",
            command=self._save_settings
        )
        self.save_settings_button.grid(row=2,column=1,padx=10,sticky=E)

        #initialise settings
        self._update_settings(self.api_key,self.host_url,self.target_database)

    def _save_settings(self,event=None):
        self._update_settings(self.api_key,self.host_url,self.target_database)
        logging.info(f"You have saved your settings: {self.target_database}: {self.host_url}")

    def _get_repository_info(self,parent) -> ttk.LabelFrame:
        frame = ttk.LabelFrame(
            master=parent,
            text="Online Repository Info"
        )
        
        preset_label = ttk.Label(
            master=frame,
            text="I am accessing: "
        )
        preset_label.grid(row=0,column=0,padx=5,pady=5,sticky=W)

        self.presets = ttk.Combobox(
            master=frame,
            values=["AE Datastore", "InvenioRDM Demo", "Custom"],
            state="readonly"
        )
        self.presets.set("AE Datastore")
        self.presets.grid(row=0, column=1, padx=5, pady=5, sticky=W)

        host_label = ttk.Label(
            master=frame,
            text="Host URL: "
        )
        host_label.grid(row=1,column=0,padx=5,pady=5,sticky=W)

        self._host_url = tk.StringVar()
        host_entry = ttk.Entry(
            master=frame,
            textvariable=self._host_url
        )
        host_entry.grid(row=1, column=1, padx=5, pady=5, sticky=W)

        #set the defaults
        self.presets.bind("<<ComboboxSelected>>",self.on_combobox_select)
        self.set_AE_db_default()

        return frame

    def _get_api_key(self,parent) -> ttk.LabelFrame:
        frame = ttk.LabelFrame(
            master=parent,
            text="API Key"
        )

        #for formatting purposes, temporary for now
        frame.grid_columnconfigure(0,minsize=250,weight=1)
        frame.grid_columnconfigure(1,minsize=25,weight=1)
        frame.grid_columnconfigure(2,minsize=25,weight=1)
        frame.grid_columnconfigure(3,minsize=25,weight=1)
        frame.grid_columnconfigure(4,minsize=25,weight=1)

        radio_label = ttk.Label(
            master=frame,
            text="I have: "
        )
        radio_label.grid(row=1, column=0, padx=5, pady=5, sticky=EW)

        self._api_radio_value = tk.StringVar(value="Option1")
        radio_option1 = ttk.Radiobutton(
            master=frame,
            text="API key",
            value="Option1",
            variable=self._api_radio_value,
            command=self.update_api_form
        )
        radio_option1.grid(row=1, column=1, padx=5, pady=5, sticky=EW)

        radio_option2 = ttk.Radiobutton(
            master=frame,
            text="API key stored in environment variable",
            value="Option2",
            variable=self._api_radio_value,
            command=self.update_api_form
        )
        radio_option2.grid(row=1, column=2, padx=5, pady=5, sticky=W)

        self.api_key_label = ttk.Label(
            master=frame,
            text="API Key: "
        )
        self.api_key_label.grid(row=2, column=0, padx=5, pady=5, sticky=W)

        self._api_key = tk.StringVar()
        api_key_entry = ttk.Entry(
            master=frame,
            textvariable=self._api_key
        )
        api_key_entry.grid(row=2, column=1,columnspan=2, pady=5, sticky=EW)

        return frame

    def update_api_form(self,event: tk.Event=None):
        if self._api_radio_value.get() == "Option1":
            self.api_key_label.configure(text="API Key: ")
        elif self._api_radio_value.get() == "Option2":
            self.api_key_label.configure(text="API Environment Variable: ")

    def on_combobox_select(self,event: tk.Event):
        """Callback for combobox selection, sets default values in some fields

        Args:
            event (Event): _description_
        """
        selected_value = event.widget.get()

        if selected_value == "AE Datastore":
            self.set_AE_db_default()
        elif selected_value == "InvenioRDM Demo":
            self.set_demo_default()
        else:
            self.set_default()

    #Sets settings for default config
    def set_AE_db_default(self):
        """Set default settings for sending to AE database
        """
        self.host_url = "https://data.ae.ic.ac.uk"

    def set_default(self):
        """No default settings
        """
        self.host_url = ""

    def set_demo_default(self):
        """Set default settings for sending to a demo instance
        """
        self.host_url = "https://inveniordm.web.cern.ch"

    @property
    def host_url(self):
        return self._host_url.get()

    @host_url.setter
    def host_url(self,value: str):
        self._host_url.set(value)

    @property
    def api_key(self):
        if self._api_radio_value.get() == "Option1":
            #if given as plain key
            return self._api_key.get()
        elif self._api_radio_value.get() == "Option2":
            #if given as environment variable
            return os.getenv(self._api_key.get(),None)
    
    @property
    def target_database(self):
        return self.presets.get()
    
    def _update_settings(self,token=None,db_url=None,db_name=None):
        if token:
            self.settings_manager.update_api_key(token)

        if db_url:
            self.settings_manager.update_db_url(db_url)

        if db_name:
            self.settings_manager.update_host_name(db_name)

    def get_settings(self) -> SettingsManager:
        return self.settings_manager


from enum import Enum
import re
import logging
from typing import Dict,Any,Type

class IdentifierType(Enum):
    """Enum denoting persistent identifier types
    """
    ORCID = "orcid"
    GND = "gnd"
    ISNI = "isni"
    ROR = "ror"

class Identifier:
    """Metadata object describing a persistent identifier associated with a person or organisation
    """
    def __init__(self, id: str, id_type: IdentifierType):
        """Class initialiser

        Args:
            id (str): ID
            id_type (IdentifierType): Persitent identifier type

        Raises:
            ValueError: _description_
        """
        self.id_type: IdentifierType = id_type

        if not self._check_valid_id(id,id_type):
            msg =f"ID {id} is not of type {id_type}"
            logging.error(msg)
            raise ValueError(msg)

        self.id = id
    
    def to_json_serialisable(self) -> Dict[str,Any]:
        """Method to serialise object as JSON

        Returns:
            Dict[str,Any]: JSON serialised object
        """
        return {
            "id": self.id,
            "id_type": self.id_type.value
        }

    @classmethod
    def from_json(cls: Type['Identifier'],data: Dict[str,Any]) -> 'Identifier':
        """Deserialise json object to reconstruct object

        Args:
            cls (Type[Identifier]): Class
            data (Dict[str,Any]): JSON serialised object

        Raises:
            ValueError: _description_

        Returns:
            Identifier: Reconstructed object
        """
        id = data["id"]
        id_type_value = data["id_type"]

        try:
            id_type = IdentifierType(id_type_value)
        except ValueError:
            msg = f"Invalid identifier type: {id_type_value}"
            logging.error(msg)
            raise ValueError(msg)

        return cls(id, id_type)  # Create and return the Identifier object
        

    def get_id_type(self) -> IdentifierType:
        """Get the id type

        Returns:
            IdentifierType: Identifier type
        """
        return self.id_type
    
    def get_id(self) -> str:
        """Get the ID

        Returns:
            str: ID
        """
        return self.id
    
    def _check_valid_id(self,id:str,id_type:IdentifierType) -> bool:
        """Check whether stated ID is valid given the ID type

        Args:
            id (str): ID
            id_type (IdentifierType): Identifier type

        Returns:
            bool: Whether id is valid
        """
        validation_methods = {
            IdentifierType.ORCID: self._is_valid_orcid_id,
            IdentifierType.GND: self._is_valid_gnd_id,
            IdentifierType.ISNI: self._is_valid_isni_id,
            IdentifierType.ROR: self._is_valid_ror_id,
        }
        
        validate = validation_methods.get(id_type)
        if validate:
            return validate(id)
        return False

    def _is_valid_orcid_id(self,id: str) -> bool:
        """Checks whether is a valid ORCID identifier

        Args:
            id (str): ID

        Returns:
            bool: Valid
        """
        #orcid id of form xxxx-xxxx-xxxx-xxxx, all numbers, last num (checksum) optionally capital 'X' for 10
        pattern = r'^\d{4}-\d{4}-\d{4}-\d{3}[\dX]$'
        if not re.match(pattern, id):
            return False

        base_digits = id.replace("-", "")[:-1]
        calculated_checksum = self._generate_check_digit_orcid(base_digits)
        return calculated_checksum == id[-1]    
    
    def _is_valid_gnd_id(self,id:str) -> bool:
        """Checks whether is a valid GND id

        Args:
            id (str): ID

        Returns:
            bool: Valid
        """
        #TODO
        return True
    
    def _is_valid_isni_id(self,id:str) -> bool:
        """Checks whether is a valid ISNI di

        Args:
            id (str): ID

        Returns:
            bool: Valid
        """

        #isni of form xxxxxxxxxxxxxxxx, all numbers (16 of them), last num (checksum) optionally capital 'X' for 10
        pattern = r'^\d{15}[\dX]$'
        if not re.match(pattern, id[:-1]):
            return False

        #TODO FIx the checksum
        #calculated_checksum = self._generate_check_digit_isni(id[:-1])
        #return calculated_checksum == id[-1]    
        return True

    def _is_valid_ror_id(self,id:str) -> bool:
        """Checks whether is a valid ROR id

        Args:
            id (str): ID

        Returns:
            bool: Valid
        """
        #TODO
        return True

    def _generate_check_digit_orcid(self,base_digits: str) -> str:
        """Generates checksum digit. Checksum code adapted from
        https://support.orcid.org/hc/en-us/articles/360006897674-Structure-of-the-ORCID-Identifier

        Args:
            base_digits (str): Base digits.

        Returns:
            str: Checksum digit
        """
        total = 0
        for digit in base_digits:
            total = (total + int(digit)) * 2

        remainder = total % 11
        result = (12 - remainder) % 11
        return "X" if result == 10 else str(result)
    
    def _generate_check_digit_isni(self,base_digits: str) -> str:
        """
        Generate the ISNI (ISO 7064 Mod 11,10) checksum digit.

        This method calculates the check digit for an ISNI (International Standard Name Identifier)
        using the ISO 7064 Mod 11,10 algorithm. The check digit is used to ensure the integrity
        of the identifier.

        Args:
            base_digits (str): The base digits of the ISNI (excluding the check digit).

        Returns:
            str: The calculated check digit, which can be a digit from '0' to '9' or 'X' if the
                check digit is 10.
        """
        weights = [2, 3, 4, 5, 6, 7, 8, 9]  # Weight factors (right to left)
        total = 0
        reversed_digits = base_digits[::-1]  # Process from right to left

        for i, digit in enumerate(reversed_digits):
            weight = weights[i % len(weights)]  # Cycle through weights
            total += int(digit) * weight

        remainder = total % 11
        check_digit = 11 - remainder

        return "X" if check_digit == 10 else str(check_digit)

    def __eq__(self, other: 'Identifier') -> bool:
        if not isinstance(other,Identifier):
            return False
        
        return (
            self.id_type == other.id_type and
            self.id == other.id
        )


import ttkbootstrap as ttk
from ttkbootstrap.constants import *
from typing import Callable, Dict
from pathlib import Path
from ttkbootstrap.scrolled import ScrolledFrame

PATH = Path(__file__).parent.parent / "assets"

class Menu(ScrolledFrame):
    """A menu frame with buttons for various actions."""

    def __init__(self, parent):
        """Initializes the Menu frame."""
        super().__init__(parent, autohide=True,width=150)
        
        self.columnconfigure(0, weight=1)

        self.button_data: Dict[str, Dict] = {
            "INFO": {"row": 0, 
                    "command": None,
                    "icon": ttk.PhotoImage(
                        name="image_icon",
                        file=PATH / "info.png"
                    )},
            "UPLOAD": {"row": 1, 
                    "command": None,
                    "icon": ttk.PhotoImage(
                        name="upload_icon",
                        file=PATH / "upload.png"
                    )},
            "REVIEW": {"row": 2, 
                    "command": None,
                    "icon": ttk.PhotoImage(
                        name="review_icon",
                        file=PATH / "review.png"
                    )},
            "EXPLORE": {"row": 3, 
                    "command": None,
                    "icon": ttk.PhotoImage(
                        name="explore_icon",
                        file=PATH / "explore.png"
                    )},
            "HELP": {"row": 4, 
                    "command": None,
                    "icon": ttk.PhotoImage(
                        name="help_icon",
                        file=PATH / "help.png"
                    )},
            "SETTINGS": {"row": 5,
                    "command": None,
                    "icon": ttk.PhotoImage(
                        name="settings_icon",
                        file=PATH / "settings.png"
                    )}
        }

        self.button_map: Dict[str, ttk.Button] = {}
        self._create_buttons()

    def _create_buttons(self) -> None:
        """Creates the menu buttons."""
        
        for text, data in self.button_data.items():
            button = ttk.Button(
                master=self,
                text=text,
                compound=TOP,
                image=data.get("icon",None),
                bootstyle=(PRIMARY,OUTLINE),
                command=data["command"]
            )
            button.grid(row=data["row"], column=0, padx=0, pady=0, sticky=NSEW)
            self.button_map[text] = button

            button.bind("<Enter>",lambda event, button=button: button.config(cursor="hand2"))
            button.bind("<Leave>",lambda event, button=button: button.config(cursor=""))

    def add_link_to_button(self, button_text: str, on_click_command: Callable) -> None:
        """Adds a click command to a button.

        Args:
            button_text (str): The text of the button to link.
            on_click_command (Callable): The function to execute on click.
        """
        if button_text in self.button_map:
            self.button_map[button_text].config(command=on_click_command)
        else:
            print(f"Warning: Button '{button_text}' not found.")


from typing import Optional, Type, List, Dict, Any
from .exceptions import HDF5ReaderException
import h5py
import logging
import os

class HDF5Reader():
    def __init__(self,filename: str, mode: str="r") -> None:
        self.filename = filename
        self.mode = mode
        self.file: Optional[h5py.File] = None
    
    def __enter__(self) -> "HDF5Reader":
        try:
            self.file = h5py.File(self.filename,self.mode)
            return self #Return HDF5Reader instance
        except Exception as e:
            logging.fatal(f"Error opening HDF5 file: {e}")
            raise e
        
    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[Type[BaseException]]) -> None:
        if self.file:
            self.file.close()
            self.file = None
    
    def get_keys(self) -> Dict[str, str]:
        """
        Recursively retrieves all keys in the HDF5 file and categorizes them as 'GROUP', 'DATASET', or 'ATTRIBUTE'.
        
        Returns:
            Dict[str, str]: A dictionary mapping keys (HDF5 paths) to their types.
        """
        if not self.file:
            msg = f"{self.filename} HDF5 file is not open."
            logging.fatal(msg)
            raise HDF5ReaderException(msg)

        key_dict = {}

        def get_keys_helper(group: h5py.Group, prefix: str = "") -> None:
            """Recursive function to traverse HDF5 groups and datasets."""
            for key in group.keys():
                full_key = f"{prefix}/{key}".lstrip("/")  # Ensure proper formatting
                
                obj = group[key]
                if isinstance(obj, h5py.Group):
                    key_dict[full_key] = "GROUP"
                    get_keys_helper(obj, full_key)  # Recurse into sub-groups
                elif isinstance(obj, h5py.Dataset):
                    key_dict[full_key] = "DATASET"

                # Add attributes (only at current level, not recursive)
                for attr_name in obj.attrs.keys():
                    attr_key = f"{full_key}/@{attr_name}"
                    key_dict[attr_key] = "ATTRIBUTE"

        get_keys_helper(self.file)
        return key_dict
    
    def get_dataset(self, dataset_key: str) -> Optional[h5py.Dataset]:
        """
        Reads a dataset from the HDF5 file.

        Args:
            dataset_key (str): The name of the dataset to read.

        Returns:
            Optional[h5py.Dataset]: The dataset object if found, otherwise None.
        """
        if not self.file:
            msg = f"{self.filename} HDF5 file is not open."
            logging.fatal(msg)
            raise HDF5ReaderException(msg)

        try:
            dataset = self.file[dataset_key]
            if isinstance(dataset, h5py.Dataset):
                return dataset
            else:
                msg = f"{dataset_key} is not a dataset."
                logging.error(msg)
                raise HDF5ReaderException(msg)
        except KeyError:
            msg = f"Dataset {dataset_key} not found in the file."
            logging.error(msg)
            raise HDF5ReaderException(msg)

    def get_attributes(self, key: str) -> Dict[str, Any]:
        """
        Retrieves attributes for a given key in the HDF5 file.

        Args:
            key (str): The key for which to retrieve attributes.

        Returns:
            Dict[str, Any]: A dictionary of attribute names and their values.
        """
        if not self.file:
            msg = f"{self.filename} HDF5 file is not open."
            logging.fatal(msg)
            raise HDF5ReaderException(msg)

        try:
            obj = self.file[key]
            return {attr_name: obj.attrs[attr_name] for attr_name in obj.attrs.keys()}
        except KeyError:
            msg = f"Key {key} not found in the file."
            logging.error(msg)
            raise HDF5ReaderException(msg)

    def dump_to_plain_file(self, target_file_name: str) -> None:
        """Convert HDF5 contents to a formatted plain-text representation."""
        if not self.file:
            raise HDF5ReaderException("HDF5 file is not open.")
        
        output_lines = self._dump_group(self.file)

        with open(target_file_name, "w") as f:
            f.write("\n".join(output_lines))
        print(f"HDF5 content saved to {target_file_name}")

    def summary(self) -> Dict[str, str]:
        if not self.file:
            raise HDF5ReaderException("HDF5 file is not open.")

        summary_dict = {
            "File Name": self.filename,
            "File Size": f"{os.path.getsize(self.filename)} bytes",
            "Groups": 0,
            "Datasets": 0,
            "Attributes": 0
        }

        def summary_helper(group: h5py.Group) -> None:
            """Recursive function to count groups, datasets, and attributes."""
            for key, obj in group.items():
                if isinstance(obj, h5py.Group):
                    summary_dict["Groups"] += 1
                    summary_helper(obj)  # Recurse into sub-groups
                elif isinstance(obj, h5py.Dataset):
                    summary_dict["Datasets"] += 1
                
                # Count attributes for both groups and datasets
                summary_dict["Attributes"] += len(obj.attrs)

        summary_helper(self.file)
        return summary_dict

    def get_dtype(self, dataset_path: str) -> Optional[str]:
        """
        Retrieves the data type of a dataset in the HDF5 file.

        Args:
            dataset_path (str): The path to the dataset.

        Returns:
            Optional[str]: The data type of the dataset if found, otherwise None.
        """
        if not self.file:
            msg = f"{self.filename} HDF5 file is not open."
            logging.fatal(msg)
            raise HDF5ReaderException(msg)

        try:
            dataset = self.file[dataset_path]
            if isinstance(dataset, h5py.Dataset):
                return str(dataset.dtype)
            else:
                msg = f"{dataset_path} is not a dataset."
                logging.error(msg)
                raise HDF5ReaderException(msg)
        except KeyError:
            msg = f"Dataset {dataset_path} not found in the file."
            logging.error(msg)
            raise HDF5ReaderException(msg)

    def get_shape(self, dataset_path: str) -> Optional[tuple]:
        """
        Retrieves the shape of a dataset in the HDF5 file.

        Args:
            dataset_path (str): The path to the dataset.

        Returns:
            Optional[tuple]: The shape of the dataset if found, otherwise None.
        """
        if not self.file:
            msg = f"{self.filename} HDF5 file is not open."
            logging.fatal(msg)
            raise HDF5ReaderException(msg)

        try:
            dataset = self.file[dataset_path]
            if isinstance(dataset, h5py.Dataset):
                return dataset.shape
            else:
                msg = f"{dataset_path} is not a dataset."
                logging.error(msg)
                raise HDF5ReaderException(msg)
        except KeyError:
            msg = f"Dataset {dataset_path} not found in the file."
            logging.error(msg)
            raise HDF5ReaderException(msg)

    def _dump_group(self, group: h5py.Group, indent: int = 0) -> List[str]:
        """Recursively process an HDF5 group and return its content as a string."""
        lines = []
        indent_str = "    " * indent

        # Group name
        lines.append(f"{indent_str} Group: {group.name}")
        if group.attrs:
            lines.append(f"{indent_str} Attributes:")
            for key, value in group.attrs.items():
                lines.append(f"{indent_str}    - {key}: {value}")

        # Iterate through datasets and groups
        for name, item in group.items():
            if isinstance(item, h5py.Group):
                lines.extend(self._dump_group(item, indent + 1))  # Recursive call
            elif isinstance(item, h5py.Dataset):
                lines.append(f"{indent_str} Dataset: {name}")
                if item.attrs:
                    lines.append(f"{indent_str} Attributes:")
                    for key, value in item.attrs.items():
                        lines.append(f"{indent_str}      - {key}: {value}")
                lines.append(f"{indent_str} Shape: {item.shape}")
                lines.append(f"{indent_str} Dtype: {item.dtype}")
                lines.append(f"{indent_str} Data: {item[...]}")

        return lines



class APIError(Exception):
    def __init__(self, message, response=None):
        super().__init__(message)
        self.response = response

class ClientError(APIError):
    pass

class ServerError(APIError):
    pass


from typing import Callable,Dict,List

class SettingsManager:
    """Manage the change of state of settings across different pages
    """
    def __init__(self,token,database_url,database_name):
        self.token = token
        self.database_url = database_url
        self.database_name = database_name

        self.update_callables: Dict[str,List[Callable]] = {
            "TOKEN": [],
            "HOST_URL": [],
            "HOST_NAME": [],
        }

    def update_api_key(self,token):
        self.token = token

        for callable in self.update_callables["TOKEN"]:
            callable()

    def update_db_url(self,database_target):
        self.database_url = database_target
        for callable in self.update_callables["HOST_URL"]:
            callable()

    def update_host_name(self,database_name):
        self.database_name = database_name
        for callable in self.update_callables["HOST_NAME"]:
            callable()

    def add_callbacks_on_update_token(self,callable: Callable):
        self.update_callables["TOKEN"].append(callable)

    def add_callbacks_on_update_host_url(self,callable: Callable):
        self.update_callables["HOST_URL"].append(callable)

    def add_callbacks_on_update_host_name(self,callable: Callable):
        self.update_callables["HOST_NAME"].append(callable)


from .validate import ValidateSession,ValidateOutput,ValidateGeometry
from typing import List
import logging

class NektarValidator:
    def __init__(self,
                session_file: str,
                geometry_file: str,
                output_file: str,
                chk_files: List[str],
                supporting_files: List[str]):
        """_summary_

        Args:
            session_file (str): _description_
            geometry_file (str): _description_
            output_file (str): _description_
            chk_files (List[str]): _description_
            supporting_files (List[str]): _description_
        """

        self.session_file: str = session_file
        self.geometry_file: str = geometry_file
        self.output_fld_file: str = output_file
        self.output_chk_files: List[str] = chk_files
        self.supporrting_files: List[str] = supporting_files 

        self.session_validator = ValidateSession(self.session_file)
        self.geometry_validator = ValidateGeometry(self.geometry_file)
        self.output_validator = ValidateOutput(self.output_fld_file)

    def validate(self) -> bool:
        """_summary_

        Returns:
            bool: _description_
        """

        self.session_validator.check_schema()
        self.geometry_validator.check_schema()
        self.output_validator.check_schema()

        return True



import os
from .hdf5_definitions import HDF5GroupDefinition,HDF5DatasetDefinition
import h5py
from .exceptions import GeometryFileException,HDF5SchemaExtraDefinitionException,HDF5SchemaExistenceException
from .exceptions import HDF5SchemaMissingDatasetException,HDF5SchemaInconsistentException,HDF5SchemaMissingDefinitionException
from typing import List,Dict,Set
from types import MappingProxyType
from NekUpload.utils import parsing

class ValidateGeometry:
    """Class responsible for all geometry validation checks
    """
    def __init__(self, file_path: str):
        """Class initialiser

        Args:
            file_path (str): Path to file
        """
        self.file = file_path
        self.file_name = os.path.basename(self.file)

    def check_schema(self) -> bool:
        """Checks whether file conforms to HDF5 geometry schema

        Raises:
            GeometryFileException: _description_

        Returns:
            bool: Passed
        """
        try:
            with h5py.File(self.file, 'r') as f:
                self.schema_checker = GeometrySchemaHDF5Validator(f)
                self.schema_checker.validate()
        except OSError as e:
            raise GeometryFileException(self.file,f"Geometry file either does not exist or is not in HDF5 format {e}")

        return True
    
class GeometrySchemaHDF5Validator:
    """Schema validator for HDF5 geometry .nekg files. Checks whether all valid groups and datasets are there.

    Raises:
        HDF5SchemaExtraDefinitionException: _description_
        HDF5SchemaExtraDefinitionException: _description_
        HDF5SchemaMissingDatasetException: _description_
        HDF5SchemaInconsistentException: _description_
        HDF5SchemaMissingDefinitionException: _description_
        HDF5SchemaMissingDefinitionException: _description_
        HDF5SchemaMissingDefinitionException: _description_
        HDF5SchemaMissingDefinitionException: _description_
    """

    NO_DIM_CONSTRAINTS = -1 #helper

    #using immutable dictionary to define what structure of each group and dataset should look like regardless of geometry file
    #dict to help associate each set with a useful descriptor, which will be beneficial later on
    BASE_GROUPS: MappingProxyType[str,HDF5GroupDefinition] = MappingProxyType({"NEKTAR": HDF5GroupDefinition("NEKTAR"),
                                            "GEOMETRY": HDF5GroupDefinition("NEKTAR/GEOMETRY",attributes=["FORMAT_VERSION"]),
                                            "MAPS": HDF5GroupDefinition("NEKTAR/GEOMETRY/MAPS"),
                                            "MESH": HDF5GroupDefinition("NEKTAR/GEOMETRY/MESH")})

    DATASETS_MANDATORY_MAPS: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType(
                                                            {"VERT": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/VERT",(NO_DIM_CONSTRAINTS,)),
                                                            "DOMAIN": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/DOMAIN",(NO_DIM_CONSTRAINTS,)),
                                                            "COMPOSITE": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/COMPOSITE",(NO_DIM_CONSTRAINTS,))
                                                            })

    DATASETS_MANDATORY_MESH: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType(
                                                                {"CURVE_NODES": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/CURVE_NODES",(NO_DIM_CONSTRAINTS,3)),
                                                                "VERT": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/VERT",(NO_DIM_CONSTRAINTS,3)),
                                                                "DOMAIN": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/DOMAIN",(NO_DIM_CONSTRAINTS,)),
                                                                "COMPOSITE": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/COMPOSITE",(NO_DIM_CONSTRAINTS,))
                                                                })
    
    DATASETS_1D_MAPS: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType(
                                                        {"SEG": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/SEG",(NO_DIM_CONSTRAINTS,)),
                                                        "CURVE_EDGE": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/CURVE_EDGE",(NO_DIM_CONSTRAINTS,))
                                                        })
    
    DATASETS_1D_MESH: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType(
                                                        {"SEG": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/SEG",(NO_DIM_CONSTRAINTS,2)),
                                                        "CURVE_EDGE": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/CURVE_EDGE",(NO_DIM_CONSTRAINTS,3))
                                                        })

    DATASETS_2D_MAPS: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType(
                                                        {"TRI": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/TRI",(NO_DIM_CONSTRAINTS,)),
                                                        "QUAD": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/QUAD",(NO_DIM_CONSTRAINTS,)),
                                                        "CURVE_FACE": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/CURVE_FACE",(NO_DIM_CONSTRAINTS,))
                                                        })
    
    DATASETS_2D_MESH: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType(
                                                        {"TRI": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/TRI",(NO_DIM_CONSTRAINTS,3)),
                                                        "QUAD": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/QUAD",(NO_DIM_CONSTRAINTS,4)),
                                                        "CURVE_FACE": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/CURVE_FACE",(NO_DIM_CONSTRAINTS,3))
                                                        })

    DATASETS_3D_MAPS: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType(
                                                        {"HEX": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/HEX",(NO_DIM_CONSTRAINTS,)),
                                                        "TET": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/TET",(NO_DIM_CONSTRAINTS,)),
                                                        "PYR": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/PYR",(NO_DIM_CONSTRAINTS,)),
                                                        "PRISM": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MAPS/PRISM",(NO_DIM_CONSTRAINTS,))
                                                        })

    DATASETS_3D_MESH: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType(
                                                        {"HEX": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/HEX",(NO_DIM_CONSTRAINTS,6)),
                                                        "TET": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/TET",(NO_DIM_CONSTRAINTS,4)),
                                                        "PYR": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/PYR",(NO_DIM_CONSTRAINTS,5)),
                                                        "PRISM": HDF5DatasetDefinition("NEKTAR/GEOMETRY/MESH/PRISM",(NO_DIM_CONSTRAINTS,5))
                                                        })
    
    DATASETS_MAPS: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType({**DATASETS_MANDATORY_MAPS,**DATASETS_1D_MAPS,
                                                                                   **DATASETS_2D_MAPS,**DATASETS_3D_MAPS})

    DATASETS_MESH: MappingProxyType[str,HDF5DatasetDefinition] = MappingProxyType({**DATASETS_MANDATORY_MESH,**DATASETS_1D_MESH,
                                                                                   **DATASETS_2D_MESH,**DATASETS_3D_MESH})

    def __init__(self,f: h5py.File):   
        """Class initialiser

        Args:
            f (h5py.File): Opened HDF5 file
        """
        self.file: h5py.File = f

        self.datasets_present: Set[str] = set()
        self.element_number: Dict[str] = {}
        
    def validate(self) -> bool:
        """Check whether the provided file conforms to the geometry HDF5 schema

        Returns:
            bool: Valid
        """
        #check mandatory groups
        for group in GeometrySchemaHDF5Validator.BASE_GROUPS.values():
            group.validate(self.file)
        
        #check all datasets
        self.datasets_present.update(self._check_mandatory_dataset(GeometrySchemaHDF5Validator.DATASETS_MANDATORY_MAPS))
        self.datasets_present.update(self._check_mandatory_dataset(GeometrySchemaHDF5Validator.DATASETS_MANDATORY_MESH))
        self.datasets_present.update(self._check_mandatory_dataset(GeometrySchemaHDF5Validator.DATASETS_1D_MAPS))
        self.datasets_present.update(self._check_mandatory_dataset(GeometrySchemaHDF5Validator.DATASETS_1D_MESH))

        self.datasets_present.update(self._check_optional_dataset(GeometrySchemaHDF5Validator.DATASETS_2D_MAPS))
        self.datasets_present.update(self._check_optional_dataset(GeometrySchemaHDF5Validator.DATASETS_2D_MESH))
        self.datasets_present.update(self._check_optional_dataset(GeometrySchemaHDF5Validator.DATASETS_3D_MAPS))
        self.datasets_present.update(self._check_optional_dataset(GeometrySchemaHDF5Validator.DATASETS_3D_MESH))

        self._check_consistent_maps_mesh_definition(self.datasets_present,GeometrySchemaHDF5Validator.DATASETS_MAPS,GeometrySchemaHDF5Validator.DATASETS_MESH)

        self.element_number = self._get_number_of_elements(self.datasets_present,GeometrySchemaHDF5Validator.DATASETS_MESH)
        self._check_element_construction(self.element_number)

        #finally check no extra unexpected payload in file
        valid_groups_keys: List[str] = [group.get_path() for group in GeometrySchemaHDF5Validator.BASE_GROUPS.values()]
        self._check_only_valid_groups_exist(valid_groups_keys)

        valid_dataset_keys: List[str] = [dataset.get_path() for dataset in GeometrySchemaHDF5Validator.DATASETS_MESH.values()] + \
                                        [dataset.get_path() for dataset in GeometrySchemaHDF5Validator.DATASETS_MAPS.values()]
        self._check_only_valid_datasets_exist(valid_dataset_keys)

        return True
    
    def _check_only_valid_groups_exist(self,valid_groups: List[str]):
        """Check that only valid HDF5 groups exist.

        Args:
            valid_groups (str): List of valid groups
        """
        #plus one to search for any extra invalid groups
        #"" is a valid group too, and is provided in function call
        valid_groups.append("")
        max_groups = len(valid_groups) + 1 
        groups = parsing.get_hdf5_groups_with_depth_limit(self.file,3,max_groups=max_groups)

        for group in groups:
            if group not in valid_groups:
                raise HDF5SchemaExtraDefinitionException(self.file,f"Encountered unkown group: {group}")

    def _check_only_valid_datasets_exist(self,valid_datasets: List[str]):
        """Check that only valid HDF5 datasets exist.

        Args:
            valid_datasets (str): List of valid datasets
        """
        max_datasets = len(valid_datasets) + 1
        datasets = parsing.get_hdf5_datasets_with_depth_limit(self.file,3,max_datasets=max_datasets)
        for dataset in datasets:
            if dataset not in valid_datasets:
                raise HDF5SchemaExtraDefinitionException(self.file,f"Encountered unkown dataset: {dataset}")

    def _check_mandatory_dataset(self,mandatory_datasets: MappingProxyType[str,HDF5DatasetDefinition]) -> Set[str]:
        """Helper function. Checks mandatory datasets and if all valid, return the keys of the present datasets

        Args:
            mandatory_datasets (MappingProxyType[str,HDF5DatasetDefinition]): Dictionary of datasets that should be present

        Returns:
            Set[str]: Set of keys denoting which datasets are present
        """
        datasets_present_key: Set[str] = set()

        for key,dataset in mandatory_datasets.items():
            if dataset.validate(self.file):
                datasets_present_key.add(key)

        return datasets_present_key

    def _check_optional_dataset(self,optional_dataset: MappingProxyType[str,HDF5DatasetDefinition]) -> Set[str]:
        """Helper function. Checks optional datasets and valid datasets will have their keys added to present datasets, which is returned.

        Args:
            optional_dataset (MappingProxyType[str,HDF5DatasetDefinition]): Dataset definitions that are optional

        Returns:
            Set[str]: Set of keys denoting which datasets are present

        Raises:
            HDF5SchemaException: _description_
        """
        datasets_present_key: Set[str] = set()

        for key,dataset in optional_dataset.items():
            try:
                dataset.validate(self.file)
                datasets_present_key.add(key)
            except HDF5SchemaExistenceException:
                pass #optional, so allow if doesn't exist, but any other definition error should be re-raised
            except Exception:
                raise

        return datasets_present_key

    def _check_consistent_maps_mesh_definition(self,
                                                present_datasets_keys: Set[str],
                                                dataset_maps: Dict[str,HDF5DatasetDefinition],
                                                dataset_mesh: Dict[str,HDF5DatasetDefinition]) -> None:
        """Check that for all present dataset keys, there is a consistent definition between the MAPS and MESH

        Args:
            present_datasets_keys (Set[str]): List of keys denoting datasets that are present
            dataset_maps (Set[str,HDF5DatasetDefinition]): Definitions of MAPS datasets
            dataset_mesh (Dict[str,HDF5DatasetDefinition]): Definitions of MESH datasets
        """
        #now check that each pair exists and have consistent shapes
        #maps can't be defined without corresponding mesh and vice versa
        for key in present_datasets_keys:
            #curve nodes only exception to above rule
            if key != "CURVE_NODES":
                self._check_pair_of_datasets(dataset_maps.get(key),dataset_mesh.get(key))

    def _check_pair_of_datasets(self, dataset_map: HDF5DatasetDefinition, dataset_mesh: HDF5DatasetDefinition) -> None:
        """Helper funcion for checking whether a map and mesh dataset have consistent definitions

        Args:
            dataset_1 (HDF5DatasetDefinition): First HDF5 dataset
            dataset_2 (HDF5DatasetDefinition): Second HDF5 dataset

        Raises:
            HDF5SchemaException: _description_
        """
        data_map = self.file.get(dataset_map.get_path())
        data_mesh = self.file.get(dataset_mesh.get_path())

        if (data_map is not None and data_mesh is None) or (data_mesh is not None and data_map is None):
            raise HDF5SchemaMissingDatasetException(self.file, f"HDF5 Schema Error: {dataset_map} and {dataset_mesh} should be defined together, \
                                                    but one exists and other doesn't")

        if data_map is not None and data_mesh is not None:
            if isinstance(data_map, h5py.Dataset) and isinstance(data_mesh, h5py.Dataset):
                shape_map = data_map.shape
                shape_mesh = data_mesh.shape
                if shape_map[0] != shape_mesh[0]:
                    raise HDF5SchemaInconsistentException(self.file, f"HDF5 Schema Error: {dataset_map} has shape {shape_map} and {dataset_mesh} \
                                                            has shape {shape_mesh}. Inconsistent lengths {shape_map[0]} != {shape_mesh[0]}")

    def _get_number_of_elements(self,
                                present_datasets_keys: Set[str],
                                dataset_mesh: Dict[str,HDF5DatasetDefinition]) -> Dict[str,int]:
        """For all datasets present in the geometry file, generate a dictionary mapping dataset keys to number of elements defined.
        Assumes consistency between maps and meshes, so meshes will be used as it contains CURVE_NODES

        Args:
            present_datasets_keys (Set[str]): Set to track number of datasets that are present
            dataset_mesh (Dict[str,HDF5DatasetDefinition]): List of valid dataset MESH definitions

        Returns:
            Dict[str,int]: Number of elements in the geometry dataset for each element type
        """

        number_elements: Dict[str,int] = {}

        for dataset_key in present_datasets_keys:
            dataset_definition: HDF5DatasetDefinition = dataset_mesh[dataset_key]
            data = self.file.get(dataset_definition.get_path())
            shape = data.shape
            elmt_num = shape[0]
            number_elements[dataset_key] = elmt_num

        return number_elements

    def _check_element_construction(self,num_elements: Dict[str,int]):
        """Make sure element construction is consistent

        Args:
            num_elements (Dict[str,int]): Mapping of HDF5 dataset to number of elements defined in that dataset

        Raises:
            HDF5SchemaMissingDefinitionException: _description_
            HDF5SchemaMissingDefinitionException: _description_
            HDF5DatasetDefinition: _description_
        """
        #3D elements can only be defined if corresponding 2D elements are present
        quads = num_elements.get("QUAD",0)
        tris = num_elements.get("TRI",0)
        
        if num_elements.get("HEX",None) and quads < 6:
            raise HDF5SchemaMissingDefinitionException(self.file,f"HDF5 Schema Error: HEX requires quads. There are only {quads} QUADS defined")

        if num_elements.get("TET",None) and tris < 4:
            raise HDF5SchemaMissingDefinitionException(self.file,f"HDF5 Schema Error: TET requires tris. There are only {tris} TRIS defined")
            
        if num_elements.get("PYR",None) and (tris < 4 or quads < 1):
            raise HDF5SchemaMissingDefinitionException(self.file,f"HDF5 Schema Error: PYR requires quads and tris. There are only {tris} TRIS and {quads} QUADS defined")
        
        if num_elements.get("PRISM",None) and (tris < 2 or quads < 4):
            raise HDF5SchemaMissingDefinitionException(self.file,f"HDF5 Schema Error: PRISM requires quads and tris. There are only {tris} TRIS and {quads} QUADS defined")



from typing import List,Any
from lxml import etree

class XMLSchemaException(Exception):
    """Exception raised for errors in the XML schema."""
    
    def __init__(self, file: str, schema_error_log: List[etree._LogEntry],message="Error in the XML schema"):        
        errors = "\n".join(f"Line {error.line}, Col {error.column}: {error.message}" for error in schema_error_log)
        full_message = f"{message} for file {file}\n{errors}" if errors else f"{message} for file {file}\n(No detailed errors found)"
        
        super().__init__(full_message)

class MissingInputFileException(Exception):
    """Exception raised when the input file is missing."""
    
    def __init__(self, file: str, message="Input file is missing"):
        full_message = f"{message}: {file}"
        super().__init__(full_message)

class MissingOutputFileException(Exception):
    """Exception raised when the output file is missing."""
    
    def __init__(self, file: str, message="Output file is missing"):
        full_message = f"{message}: {file}"
        super().__init__(full_message)

class HDF5SchemaException(Exception):
    """Exception raised for errors in the HDF5 schema."""
    
    def __init__(self, file: str, message="HDF5 schema error"):
        full_message = f"{message}: {file}"
        super().__init__(full_message)

class HDF5SchemaExistenceException(HDF5SchemaException):
    """Exception raised for errors in the HDF5 schema."""
    pass

class HDF5SchemaMissingDatasetException(HDF5SchemaException):
    """Exception raised for errors in the HDF5 schema."""
    pass

class HDF5SchemaInconsistentException(HDF5SchemaException):
    """Exception raised for errors in the HDF5 schema."""
    pass

class HDF5SchemaMissingDefinitionException(HDF5SchemaException):
    """Exception raised for errors in the HDF5 schema."""
    pass

class HDF5SchemaExtraDefinitionException(HDF5SchemaException):
    """Exception raised for errors in the HDF5 schema."""
    pass

class GeometryFileException(Exception):
    def __init__(self, file: str, message="Geometry file error"):
        full_message = f"{message}: {file}"
        super().__init__(full_message)

class OutputFileException(Exception):
    def __init__(self, file: str, message="Geometry file error"):
        full_message = f"{message}: {file}"
        super().__init__(full_message)


import ttkbootstrap as ttk
from ttkbootstrap.constants import *
import tkinter as tk
from .create_author_window import CreateAuthorOrgWindow,CreateAuthorPersonWindow
from typing import Dict,List,Any
from NekUpload.frontend.components.settings_manager import SettingsManager
from NekUpload.frontend.components.scrollbox import ScrolledListbox
class UploadInfoFrame(ttk.Labelframe):
    def __init__(self,root,parent,setting_manager: SettingsManager):
        super().__init__(
            master=parent,
            text="Basic Upload Information",
            bootstyle=PRIMARY,
            padding=10
        )
        self.root = root
        self.setting_manager = setting_manager

        self.columnconfigure(0,weight=1)
        self.columnconfigure(1,weight=1)
        self.columnconfigure(2,weight=5)
        self.columnconfigure(3,weight=5)
        self.rowconfigure(0,weight=1)

        community_upload_label = ttk.Label(
            master=self,
            text="Upload to community: ",
            bootstyle=PRIMARY
        )
        community_upload_label.grid(row=0,column=0,sticky=W)

        self.presets = ttk.Combobox(
            master=self,
            state="readonly"
        )
        self.presets.grid(row=0, column=1, padx=5, pady=5, sticky=EW)
        self.presets.bind("<<ComboboxSelected>>",self._update_community_slug_value)
        self.setting_manager.add_callbacks_on_update_host_name(self._update_community_slug_value)

        community_slug_label = ttk.Label(
            master=self,
            text="Community (URL Slug or UUID): ",
            bootstyle=PRIMARY
        )
        community_slug_label.grid(row=1,column=0,sticky=W)

        self._community_slug = tk.StringVar()
        self._community_slug.set("nektar")
        self.community_slug_entry = ttk.Entry(
            master=self,
            textvariable=self._community_slug
        )
        self.community_slug_entry.grid(row=1,column=1,padx=5,pady=5,sticky=EW)

        publication_date_label = ttk.Label(
            master=self,
            text="Publication Date: ",
            bootstyle=PRIMARY
        )
        publication_date_label.grid(row=2,column=0,sticky=EW)

        self.publication_date = tk.StringVar()
        self.date_entry = ttk.DateEntry(
            master=self,
            dateformat=r"%Y-%m-%d"
        )
        self.date_entry.grid(row=2,column=1,columnspan=1,padx=5,pady=5,sticky=EW)

        add_author_frame: ttk.Frame = self._add_authors_frame(self)
        add_author_frame.grid(row=3,column=0,sticky=NSEW,rowspan=2,columnspan=3)

        #update final value of combobox
        self._update_community_slug_value()

    def _add_authors_frame(self,parent) -> ttk.Frame:
        frame = ttk.Frame(
            master=parent
        )
        frame.columnconfigure(0,weight=1)
        frame.columnconfigure(1,weight=1)
        frame.columnconfigure(2,weight=1)

        #note that this mirrors self.author_listbox data in terms of order
        #should probably consolidate at some point
        self.authors = []

        add_author_person_button = ttk.Button(
            master=frame,
            text="Add Person",
            command=self._create_author_person
        )
        add_author_person_button.grid(row=1,column=0,padx=5,pady=5,sticky=NSEW)

        add_author_org_button = ttk.Button(
            master=frame,
            text="Add Organisation",
            command=self._create_author_org
        )
        add_author_org_button.grid(row=1,column=1,padx=5,pady=5,sticky=NSEW)

        #have a listbox to specify the creation
        self.author_listbox_frame = ScrolledListbox(frame)
        self.author_listbox = self.author_listbox_frame.listbox
        self.author_listbox_frame.grid(row=2,column=0,columnspan=2,rowspan=2,sticky=(NSEW))

        #add a delete button
        delete_button = ttk.Button(frame,text="Delete",command=self._delete_selected_author)
        delete_button.grid(row=4,column=1,padx=5,pady=5,sticky=NSEW)
        return frame

    def _delete_selected_author(self) -> None:
        """Delete the selected author from the list of authors. Deletion is accomplished via a Button connected to a Listbox.
        """
        selection_indices = self.author_listbox.curselection()
        if selection_indices:
            # 1. Get the items to delete *before* modifying the listbox
            items_to_delete = [self.author_listbox.get(index) for index in selection_indices]

            # 2. Delete from the listbox (reverse order to prevent issues with shifting indices)
            for index,item in zip(sorted(selection_indices, reverse=True),sorted(items_to_delete,reverse=True)):
                self.author_listbox.delete(index)

                #delete from self.authors using the index
                # self.authors should mirror same order as listbox
                try:
                    print(f"Removed {self.authors.pop(index)}")
                except ValueError:
                    print(f"Warning: Item '{item}' not found in self.authors")  # Handle potential mismatch

        print(self.authors)

    def _create_author_person(self) -> None:
        """Opens a new window with a form to specify the author as a person
        """
        
        # The lambda function will be executed later when triggered by an event.
        # By that time, new_window will have already been created by __init__.
        # This works because lambdas capture variables by reference, not by value.
        new_window = CreateAuthorPersonWindow(self.root,lambda: self.submit_author_person_info(new_window))

    def _create_author_org(self) -> None:
        """Opens a new window with a form to specify the author as an organisation
        """
        # The lambda function will be executed later when triggered by an event.
        # By that time, new_window will have already been created by __init__.
        # This works because lambdas capture variables by reference, not by value.
        new_window = CreateAuthorOrgWindow(self.root,lambda: self.submit_author_org_info(new_window))

    def submit_author_person_info(self,window: CreateAuthorPersonWindow):
        """Takes person data from CreateAuthorPersonWindow and stores it. Closes window on success.

        Args:
            window (CreateAuthorPersonWindow): Window containing form to specify user info
        """
        author_data = {}
        author_data['type'] = 'personal'
        author_data['given_name'] = window.given_name
        author_data['last_name'] = window.last_name
        author_data['name'] = f"{author_data['given_name']} {author_data['last_name']}"
        author_data['affiliation'] = window.affiliation
        author_data['id_type'] = window.id_type
        author_data['id'] = window.id

        print("Author Data: ", author_data)

        self.authors.append(author_data)        
        self.author_listbox.insert(END,f"{author_data['name']}")

        window.destroy()

    def submit_author_org_info(self,window: CreateAuthorOrgWindow):
        """Takes person data from CreateAuthorOrgWindow and stores it. Closes window on success.

        Args:
            window (CreateAuthorOrgWindow): Window containing form to specify organisation info
        """
        author_data = {}
        author_data['type'] = 'organizational'
        author_data['name'] = window.name
        author_data['affiliation'] = window.affiliation
        author_data['id_type'] = window.id_type
        author_data['id'] = window.id

        print("Author Data: ", author_data)
        self.authors.append(author_data)        
        self.author_listbox.insert(END,f"{author_data['name']}")

        window.destroy()
    
    def _update_community_slug_value(self,event: tk.Event=None):
        """Callback for combobox selection, sets default values in some fields

        Args:
            event (Event): _description_
        """
        #update options based on database url
        database_name: str = self.setting_manager.database_name        
        if database_name == "AE Datastore":
            self.presets.configure(values=["Nektar++","Custom"])
        elif database_name == "InvenioRDM Demo":
            self.presets.configure(values=["NekUpload Demo","Custom"])
        else:
            self.presets.configure(values=["Custom"])
        
        if event:
            selected_value = event.widget.get()
        else:
            selected_value = self.presets.get()

        if database_name == "AE Datastore":
            if selected_value == "Nektar++":
                self._community_slug.set("nektar")#tbc
            else:
                self._community_slug.set("")
        elif database_name == "InvenioRDM Demo":
            if selected_value == "NekUpload Demo":
                self._community_slug.set("test_nekupload")
            else:
                self._community_slug.set("")
        else:
            self._community_slug.set("")

        #if not an event, then that means switched due to settings
        #hence, set a default config
        #brute force for now
        if not event:
            if database_name == "AE Datastore":
                self.presets.set("Nektar++")
                self._community_slug.set("nektar")
            elif database_name == "InvenioRDM Demo":
                self.presets.set("InvenioRDM Demo")
                self._community_slug.set("test_nekupload")
            else:
                self.presets.set("Custom")
                self._community_slug.set("")

    #Sets settings for default config
    def set_AE_db_default(self):
        """Set default settings for sending to AE database
        """
        self.host_url = "https://data.ae.ic.ac.uk"

    def set_default(self):
        """No default settings
        """
        self.host_url = ""

    @property
    def author_list(self) -> List[Dict[str,Any]]:
        """Read only access to list of authors

        Returns:
            List[Dict[str,Any]]: Dictionary containing information on authors
        """
        return self.authors
    
    @property
    def community_slug(self) -> str:
        return self._community_slug.get()
    
    @property
    def publication_date_iso(self) -> str:
        """Return in  format YYYY-MM-DD

        Returns:
            str: _description_
        """
        return self.date_entry.entry.get()


import click
from NekUpload.upload import InvenioRDM
from NekUpload.metadata import *
from NekUpload.metadata.user import InvenioUserInfo
from typing import Dict,List,Tuple,Any
from dataclasses import dataclass,field
from datetime import date
import json
import os
import pathlib
import yaml
from ruamel.yaml import YAML
from io import StringIO
from NekUpload.utils.xml_reader import XMLReader

@dataclass
class Config:
    #use field to ensure independent instances of authors
    authors: List[InvenioOrgInfo | InvenioPersonInfo] = field(default_factory=list)
    metadata: InvenioMetadata = None

    CONTEXT_FILE: pathlib.Path = "config.json"

    def to_json(self):
        data = {
            "CONTEXT_FILE": str(self.CONTEXT_FILE)
        }
        
        #only jsonify non empty data
        if self.metadata:
            data["metadata"] = self.metadata.to_json_serialisable()
        
        if self.authors:
            data["authors"] = [author.to_json_serialisable() for author in self.authors]

        return data
    
    @classmethod
    def from_json(cls,data: Dict[str,Any]) -> 'Config':
        config = Config()
        
        config.CONTEXT_FILE = pathlib.Path(data["CONTEXT_FILE"])

        if metadata := data.get("metadata",None):
            config.metadata = InvenioMetadata.from_json(metadata)

        if authors := data.get("authors",[]):
            config.authors = [InvenioUserInfo.from_json(author) for author in authors]

        return config

    @classmethod
    def from_yaml(cls,data: Dict[str,Any]) -> 'Config':
        metadata: Dict[str,Any] = data["metadata"]
        title = metadata["title"]
        publication_date = metadata["publication_date"]
        description = metadata.get("description",None)
        publisher = metadata.get("publisher","NekUpload-CLI")
        authors: List[Dict[str,Any]] = metadata["authors"] 

        def get_authors(yaml_data: Dict[str,Any]) -> InvenioUserInfo:
            """Turn yaml representation of authors into json"""
            json = yaml_data

            identifiers = json.get("identifiers",[])

            identifier = None
            if identifiers:
                #if identifiers have been specified, remove it from json
                #redefine it
                json.pop("identifiers")
                
                #right now, only orcid is supported
                orcid = identifiers.get("orcid")
                if orcid:
                    identifier = Identifier(orcid,IdentifierType.ORCID)

            #define author
            author = InvenioUserInfo.from_json(json)

            if identifier:
                author.add_identifier(identifier)
            return author

        author_list = [get_authors(author) for author in authors]

        config = Config()
        config.authors = author_list
        metadata_obj = InvenioMetadata(title,publication_date,author_list)        
        metadata_obj.add_publisher(publisher)

        if description:
            #currently no support for description, but will change in future
            pass

        config.metadata = metadata_obj
        return config

@click.group()
@click.option("--config","-c",type=click.Path(dir_okay=False, path_type=pathlib.Path),default="config.json",help="Use specified config file, defaults to config.json")
@click.pass_context
def cli(ctx: click.Context, config: pathlib.Path):
    ctx.ensure_object(Config)
    try:
        with open(config,"r") as f:
            ctx.obj = Config.from_json(json.load(f))
    except FileNotFoundError:
        ctx.obj = Config()
    except json.JSONDecodeError:
        click.echo(f"Warning: Could not parse {config}. Starting with a fresh config.")
        ctx.obj = Config()

    ctx.obj.CONTEXT_FILE = config
    ctx.call_on_close(lambda: save_config(ctx,config))

def save_config(ctx: click.Context,config_file: pathlib.Path):
    with open(config_file, "w") as f:
        json.dump(ctx.obj.to_json(), f, indent=4)
    click.echo(f"Config saved to {config_file}")

@cli.command()
@click.argument('given_name')
@click.argument('last_name')
@click.option('--orcid',help="orcid identifier")
@click.option('--affiliation',help="affiliated with what organisation")
@click.pass_context
def add_author_person(ctx: click.Context,given_name: str,last_name: str,orcid: str=None,affiliation: str=None):
    """Add an author, who is a person

    \b
    GIVEN_NAME: Given name of author (use quotes if there is a space)
    LAST_NAME: Last name of author (use quotes if there is a space)
    """
    author = InvenioPersonInfo(given_name,last_name)

    if orcid:
        author.add_identifier(Identifier(orcid,IdentifierType.ORCID))

    config: Config = ctx.obj
    config.authors.append(author)

@cli.command()
@click.argument('name')
@click.option('--orcid',help="orcid identifier")
@click.option('--affiliation',help="affiliated with what organisation")
@click.pass_context
def add_author_org(ctx: click.Context,name: str,orcid: str=None,affiliation: str=None):
    """Add an author, who is a organisation

    \b
    NAME: Name of organisation (use quotes if there is a space)
    LAST_NAME: Last name of author (use quotes if there is a space)
    """
    author = InvenioOrgInfo(name)

    if orcid:
        author.add_identifier(Identifier(orcid,IdentifierType.ORCID))

    config: Config = ctx.obj
    config.authors.append(author)

@cli.command()
@click.argument("title")
@click.option("--pub-date",help="Publication date, defaults to today's date")
@click.pass_context
def add_info(ctx: click.Context,title: str,pub_date: str=None):
    """Add information about datasets to be uploaded

        \b
        TITLE: dataset title
    """
    config: Config = ctx.obj

    if not pub_date:
        today: date = date.today() 
        pub_date = today.isoformat()

    metadata = InvenioMetadata(title,pub_date,config.authors)
    config.metadata = metadata

@cli.command()
@click.option('-u',"--user-config",help="user-defined yaml file containing upload settings")
@click.option("--api-key", envvar="NEKTAR_DB_API_KEY", help="Your API key (or set environment variable NEKTAR_DB_API_KEY)")
@click.option("--host", envvar="NEKTAR_DB_HOST", help="Host name of database (or set environment variable NEKTAR_DB_HOST)")
@click.option("--community-slug", envvar="NEKTAR_DB_COMMUNITY", help="Community id to upload to (or set environment variable NEKTAR_DB_COMMUNITY)")
@click.option("-dir","--directory",type=click.Path(exists=True), help="Directory containing files to be uploaded")
@click.option('-f', '--file', multiple=True, type=click.Path(exists=True), help="Path to a file to be uploaded, can specify multiple e.g. -f file1 -f file2")
@click.pass_context
def upload(ctx: click.Context,api_key: str=None, host: str=None, community_slug: str=None,directory: str=None,file:Tuple[str]=None,user_config:str=None):
    """Validate and upload the files to the specified database
    """

    #for this, all data is assumed to be within the user config file
    if user_config:
        upload_user_config(user_config)
        return

    if not api_key:
        click.echo("Error: API key is required. Set the NEKTAR_DB_API_KEY environment variable or use the --api-key option.")
        return
    
    if not host:
        click.echo("Error: Host name is required. Set the NEKTAR_DB_HOST environment variable or use the --host option")
        return

    if not community_slug:
        click.echo("Error: Community id is required. Set the NEKTAR_DB_COMMUNITY environment variable or use the --community-slug option")
        return 
    
    if not directory and not file:
        click.echo("Error: No files detected. Set files either via -f or the directory via -dir") 

    files = None
    if directory:
        files = [os.path.join(directory, f) for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f))]
    elif file:
        files = [f for f in file]
    
    config: Config = ctx.obj
    
    if not config.metadata:
        click.echo("Error: No metadata set. Please use the command add_info")
        return 
    
    if config.authors == []:
        click.echo("Error: No authors set. Please set an author with command add_author_person or add_author_org")
        return 
    
    db = InvenioRDM()
    metadata = {"metadata": config.metadata.get_metadata_payload()}
    db.upload_files(host,api_key,files,metadata,community_slug)

def upload_user_config(user_config_file: str):
    yaml: Dict[str,Any] = read_yaml_file(user_config_file)
    config = Config.from_yaml(yaml)

    db = yaml["database"]
    host_name = db["host_name"]
    #unsafe key storage atm
    api_key = db["api_key"]
    community_slug = db["community_slug"]

    upload_data = yaml["upload"]
    upload_files = upload_data.get("files",[])
    upload_dirs = upload_data.get("directories",[])

    #collect all files specified in directory
    if upload_dirs:
        for dir in upload_dirs:
            files = [os.path.join(dir, f) for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]
            upload_files.append(files)

    db = InvenioRDM()
    metadata = {"metadata": config.metadata.get_metadata_payload()}
    
    db.upload_files(host_name,api_key,upload_files,metadata,community_slug)


@cli.command()
@click.pass_context
def list_authors(ctx: click.Context):
    """List all authors."""
    config: Config = ctx.obj
    for author in config.authors:
        click.echo(author)

def read_yaml_file(filepath):
    """Reads a YAML file and returns the parsed data.

    Args:
        filepath: The path to the YAML file.

    Returns:
        The parsed YAML data as a Python object (usually a dictionary or list),
        or None if an error occurs (e.g., file not found, invalid YAML).
    """
    try:
        with open(filepath, 'r') as file:
            yaml_data = yaml.safe_load(file)  # Use safe_load to avoid potential security issues
            return yaml_data
    except FileNotFoundError:
        print(f"Error: File not found at {filepath}")
        return None
    except yaml.YAMLError as e:
        print(f"Error parsing YAML: {e}")
        return None

@cli.command()
@click.option("--name", "-n", help="Create a yaml configuration template for user with specified name", default="config.yaml")
def quickstart(name: str):
    """Creates a template configuration file for nekupload"""

    today = date.today().isoformat()

    yaml_data = {
        'metadata': {
            'title': '<TITLE HERE>',
            'publication_date': f'{today}',
            'authors': [
                {
                    'type': 'personal',
                    'given_name': '<NAME>',
                    'family_name': '<SURNAME>',
                    'identifiers': {
                        'orcid': 'xxxx-xxxx-xxxx-xxxx'
                    }
                },
                {
                    'type': 'organizational',
                    'name': 'Imperial College London'
                }
            ],
            'description': 'This is the description'
        },
        'upload': {
            'files': [
                'YOUR FILES HERE',
                'YOUR FILES HERE'
            ],
            'directories': [
                'DIRECTORY CONTAINING FILES'
            ]
        },
        'database': {
            'host_name': '<YOUR_HOST_NAME_HERE>',
            'api_key': '<YOUR_API_KEY_HERE>',
            'community_slug': '<YOUR_COMMUNITY_SLUG>'
        }
    }

    yaml_obj = YAML()
    yaml_obj.preserve_quotes = True

    # Ensure proper indentation

    # God (GPT) knows what happens after this point for commenting 
    # Convert to CommentedMap using StringIO
    stream = StringIO()
    yaml_obj.dump(yaml_data, stream)
    yaml_data = yaml_obj.load(stream.getvalue())

    # Add comments
    yaml_data.yaml_set_comment_before_after_key('metadata', before='This is a template with minimum required descriptors and some key optionals')
    yaml_data.yaml_set_comment_before_after_key('metadata', before='Mandatory fields are denoted with <>')
    yaml_data.yaml_set_comment_before_after_key('metadata', before='Refer to documentation for more detail')

    metadata = yaml_data['metadata']
    metadata.yaml_set_comment_before_after_key('publication_date', before='Defaults to today, of form "YYYY-MM-DD" in string format')
    metadata.yaml_set_comment_before_after_key('authors', before='Can have multiple person or oraganisation as authors, edit/delete as neccesary')

    metadata["authors"][0].yaml_set_comment_before_after_key("identifiers",before="Identifiers are optional, only orcid is supported")
    metadata["authors"][0].yaml_set_comment_before_after_key("identifiers",before="Delete identifiers if not needed")

    yaml_data.yaml_set_comment_before_after_key('upload', before='Specify files and directories containing files to upload here')
    yaml_data.yaml_set_comment_before_after_key('upload', before='Edit/delete as necessary')

    yaml_data.yaml_set_comment_before_after_key('database', before='Information for InvenioRDM database connection')

    

    with open(name, "w") as f:
        yaml_obj.dump(yaml_data, f)

@cli.command()
@click.argument("file1")
@click.argument("file2")
@click.argument("merge_file")
def mergeXML(file1: str,file2: str, merge_file: str):
    """Merges two XML files. If repeated first-level elements are detected,
    elements from file2 will replace those in file1. Otherwise, will be appended.

    \b
    file1 (str): Path to XML file
    file2 (str): Path to XML file
    merge_file (str): Name of file where merged XML file is written to
    """
    with XMLReader(file1) as f:
        f.merge_first_level_elements_with(file2,merge_file)

def main():#Entry point
    cli()

if __name__ == "__main__":
    main()


import ttkbootstrap as ttk
from ttkbootstrap.constants import *
import tkinter as tk
from tkinter import filedialog
from NekUpload.frontend.components.help import HelpNotification
from NekUpload.frontend import style_guide
from ttkbootstrap.scrolled import ScrolledText
from NekUpload.frontend.components.scrollbox import ScrolledListbox
from typing import List

class UploadGeometryFrame(ttk.LabelFrame):
    def __init__(self,parent):
        super().__init__(
            master=parent,
            text="Geometry",
            bootstyle=DANGER,
            padding=10
            )

        self.columnconfigure(0,weight=1)
        self.columnconfigure(1,weight=1)
        self.rowconfigure(0,weight=1)
        self.rowconfigure(1,weight=1)

        label = ttk.Label(
            master=self,
            text="Upload all geometry related files here.",
            font=("TKDefaultFont", 12),
            anchor="w",
            justify="left",
        )
        label.grid(row=0,column=0,columnspan=2,pady=2,sticky=(NSEW))

        n = ttk.Notebook(self,bootstyle=SECONDARY)
        n.grid(row=1,column=0,columnspan=2,sticky=NSEW)
        
        f1: ttk.Frame = self._upload_geometry(n)
        f2: ttk.Frame = self._link_geometry_to_existing_record(n)

        n.add(f1,text="Upload from Source")
        n.add(f2,text="Link to Existing Record")
    
    def _link_geometry_to_existing_record(self,parent) -> ttk.Frame:
        frame = ttk.Frame(master=parent)

        todo = ttk.Label(master=frame,
                        text="TBC",
                        bootstyle=DANGER,
                        anchor="w")
        todo.grid(row=0,column=0,sticky=NSEW)

        return frame
    
    def _upload_geometry(self,parent) -> ttk.Frame:
        frame = ttk.Frame(master=parent)

        frame.rowconfigure(0,weight=1)
        frame.rowconfigure(1,weight=1)
        frame.columnconfigure(0,weight=1)
        frame.columnconfigure(1,weight=1)

        geometry_file_frame: ttk.Labelframe = self._geometry_mandatory_info(frame)
        geometry_file_frame.grid(row=0,column=0,sticky=NSEW)
        geometry_optional_data_frame = self._optional_geometry_upload_frame(frame)
        geometry_optional_data_frame.grid(row=1,column=0,columnspan=2,sticky=NSEW)

        return frame
    
    def _geometry_mandatory_info(self,parent) -> ttk.Labelframe:
        geometry_file_frame= ttk.Labelframe(
            master=parent,
            text="Mandatory"
        )
        geometry_file_frame.columnconfigure(0,weight=1)
        geometry_file_frame.columnconfigure(1,weight=3)
        geometry_file_frame.columnconfigure(2,weight=1)
        geometry_file_frame.columnconfigure(3,weight=1)

        ################################
        # Ask for title
        title_label = ttk.Label(
            master=geometry_file_frame,
            text="Geometry Title: "
        )
        title_label.grid(row=0,column=0,sticky=NSEW)
        
        self._geometry_title = tk.StringVar()
        self.geometry_title_entry = ttk.Entry(
            master=geometry_file_frame,
            textvariable=self._geometry_title,
        )
        self.geometry_title_entry.grid(row=0,column=1,sticky=NSEW)
        
        #for highlighting missing fields and turning off red when user is inputting
        self.geometry_title_entry.bind("<FocusOut>",lambda event: style_guide.highlight_mandatory_entry_on_focus_out(self.geometry_title_entry))
        self.geometry_title_entry.bind("<FocusIn>",lambda event: style_guide.highlight_mandatory_entry_on_focus_in(self.geometry_title_entry))

        ################################
        # Ask for geometry file
        geometry_file_label = ttk.Label(
            master=geometry_file_frame,
            text="Select Geometry File: "
        )
        geometry_file_label.grid(row=1,column=0,sticky=NSEW)

        self._geometry_file = tk.StringVar()
        self.geometry_file_entry = ttk.Entry(
            master=geometry_file_frame,
            textvariable=self._geometry_file,
        )
        self.geometry_file_entry.grid(row=1,column=1,sticky=NSEW)

        #for highlighting missing fields and turning off red when user is inputting
        self.geometry_file_entry.bind("<FocusOut>",lambda event: style_guide.highlight_mandatory_entry_on_focus_out(self.geometry_file_entry))
        self.geometry_file_entry.bind("<FocusIn>",lambda event: style_guide.highlight_mandatory_entry_on_focus_in(self.geometry_file_entry))


        def browse_file():
            file_path = filedialog.askopenfilename(
                title="Select Geometry File",
                filetypes=(("Geometry Files",".nekg"),)
            )
            self._geometry_file.set(file_path)

        browse = ttk.Button(
            master=geometry_file_frame,
            text="Browse Files",
            command=browse_file
        )
        browse.grid(row=1,column=2,sticky=NSEW)

        help_logo:ttk.Label = HelpNotification(geometry_file_frame)
        help_logo.add_help_message("This will be the title of the geometry record")
        help_logo2:ttk.Label = HelpNotification(geometry_file_frame)
        help_logo2.add_help_message("Only HDF5 formatted geometry files are accepted")
        help_logo.grid(row=0, column=3, sticky=NSEW)
        help_logo2.grid(row=1, column=3, sticky=NSEW)

        return geometry_file_frame
    
    #OPTIONAL INFO
    def _optional_geometry_upload_frame(self,parent) -> ttk.Labelframe:
        frame = ttk.Labelframe(
            master=parent,
            bootstyle=DEFAULT,
            text="Optional"
        )

        frame.columnconfigure(0,weight=1)
        frame.columnconfigure(1,weight=1)
        frame.columnconfigure(2,weight=1)
        frame.columnconfigure(3,weight=1)
        frame.columnconfigure(4,weight=1)
        frame.columnconfigure(5,weight=1)

        label = ttk.Label(
            master=frame,
            bootstyle=DEFAULT,
            text="Description: ",
            anchor="w"
        )
        label.grid(row=0,column=0,sticky=EW)

        self._description_text = ScrolledText(
            master=frame,
            autohide=True,
            bootstyle=DEFAULT
        )
        self._description_text.text.configure(height=10)
        self._description_text.grid(row=1,column=0,columnspan=3,sticky=NSEW)

        optional_files = ttk.Label(
            master=frame,
            bootstyle=DEFAULT,
            text="Optional Files: ",
            anchor="w"
        )
        optional_files.grid(row=0,column=3,sticky=EW)

        optional_files_button = ttk.Button(
            master=frame,
            bootstyle=SUCCESS,
            text="Browse Files",
            command=self._select_files_listbox
        )
        optional_files_button.grid(row=0,column=4,padx=5,sticky=NSEW)

        optional_files_delete_button = ttk.Button(
            master=frame,
            bootstyle=(DANGER),
            text="Delete Files",
            command=self._delete_files_listbox
        )
        optional_files_delete_button.grid(row=0,column=5,padx=5,sticky=NSEW)

        #have a listbox to specify the creation
        self.optional_files_listbox_frame = ScrolledListbox(frame)
        self.optional_files_listbox = self.optional_files_listbox_frame.listbox
        self.optional_files_listbox_frame.grid(row=1,column=3,columnspan=3,rowspan=2,sticky=NSEW)

        return frame

    @property
    def geometry_file_name(self):
        return self._geometry_file.get()
    
    @property
    def geometry_dataset_title(self):
        return self._geometry_title.get()
    
    @property
    def geometry_description(self):
        return self._description_text.text.get("1.0", "end-1c")
    
    @property
    def geometry_optional_files(self):
        return [self.optional_files_listbox.get(i) for i in range(self.optional_files_listbox.size())]
    
    def add_error_style_to_mandatory_entries(self):
        if not self.geometry_file_entry.get():
            style_guide.show_error_in_entry(self.geometry_file_entry)

        if not self.geometry_title_entry.get():
            style_guide.show_error_in_entry(self.geometry_title_entry)

    def _select_files_listbox(self) -> None:
        """Add files selected from filedialogue into listbox, ensuring no duplication
        """
        filetype = ("All Files","*")
        selected_files = filedialog.askopenfilenames(title="Select Files", filetypes=(filetype,))

        if selected_files:
            existing_files = set(self.geometry_optional_files)  # Use a set for efficient lookup

            for file in selected_files:
                if file not in existing_files:  # Check for duplicates
                    self.optional_files_listbox.insert(END, file)
                    existing_files.add(file)  # Keep the set updated
                else:
                    print(f"Duplicate file: {file} - not added.")

            print(f"Files: {self.geometry_optional_files}")  # Print updated file list
        else:
            print("No Files Selected")

    def _delete_files_listbox(self) -> None:
        """Delete selected files in the listbox
        """
        selection_indices = self.optional_files_listbox.curselection()

        if selection_indices:
            # 1. Get the items to delete *before* modifying the listbox
            items_to_delete = [self.optional_files_listbox.get(index) for index in selection_indices]

            # 2. Delete from the listbox (reverse order to prevent issues with shifting indices)
            for index,item in zip(sorted(selection_indices, reverse=True),sorted(items_to_delete,reverse=True)):
                self.optional_files_listbox.delete(index)

            print(f"Deleted files. Remaining: {self.geometry_optional_files}")



import ttkbootstrap as ttk
from ttkbootstrap.constants import *

class Header(ttk.Frame):
    def __init__(self,parent):
        super().__init__(parent)

        self.configure(bootstyle=PRIMARY)
        self.columnconfigure(0,weight=1)

        logo_text = ttk.Label(
            master = self,
            text = "NekUpload",
            font = ("TkDefaultFixed",30),
            bootstyle=(INVERSE, PRIMARY)
        )

        logo_text.grid(row=0,column=0,pady=5)

        description = ttk.Label(
            master = self,
            text = "NekUpload is a utility tool that handles the upload and review process of Nektar++ datasets to AE Datastore",
            font = ("TkDefaultFixed",15),
            bootstyle=(INVERSE, PRIMARY)
        )

        description.grid(row=1,column=0,pady=10)


import tkinter as tk
import ttkbootstrap as ttk
from ttkbootstrap.scrolled import ScrolledText
from ttkbootstrap.constants import *

import logging

class TerminalHandler(logging.Handler):
    """Custom logging handler

    Args:
        logging (_type_): _description_
    """
    def __init__(self, terminal: 'TerminalWidget'):
        super().__init__()
        self.terminal = terminal

    def emit(self, record):
        """Writes a log to the terminal

        Args:
            record (_type_): _description_
        """
        msg = self.format(record)
        level = record.levelname  # Get the log level

        tag = None  # Default tag
        if level == "ERROR":  # Check for ERROR level
            tag = "error"  # Use the "error" tag

        self.terminal.write(msg + "\n", tag=tag)  # Pass the tag to write()
        self.terminal.terminal.update_idletasks()

class TerminalWidget(ttk.Frame):
    """Terminal widget to emulate a terminal

    Args:
        ttk (_type_): _description_
    """
    def __init__(self, parent, height=10, **kwargs):
        """Creates a widget containing a text terminal

        Args:
            parent (ttk.Frame): Parent frame
            height (int, optional): Height of terminal. Defaults to 10.
        """
        super().__init__(parent, **kwargs)

        self.terminal = ScrolledText(
            master=self, 
            wrap=tk.WORD,
            state=tk.DISABLED,
            height=height,
            bootstyle=INFO,
            autohide=True
        )
        self.terminal.pack(fill=tk.BOTH, expand=True)

        # Configure the "error" tag for red text
        self.terminal.text.tag_configure("error", foreground="red")  # Set foreground to red

    def write(self, text: str, tag=None):
        """Appends text to the terminal.

        Args:
            text: The text to add.
            tag: An optional tag to apply to the text (for styling).
        """
        self.terminal.text.config(state=tk.NORMAL)
        self.terminal.text.insert(tk.END, text, tag)  # Use the tag here
        self.terminal.text.see(tk.END)
        self.terminal.text.config(state=tk.DISABLED)

    def clear(self):
        """Clears the terminal content."""
        self.terminal.text.config(state=tk.NORMAL)
        self.terminal.text.delete(1.0, tk.END)  # Delete from beginning to end
        self.terminal.text.config(state=tk.DISABLED)

    def tag_configure(self, tag, **kwargs):
        """Configures a tag for styling text."""
        self.terminal.text.tag_configure(tag, **kwargs)


import h5py
import numpy as np
from typing import Tuple,List,Dict

class HDF5Extractor:
    """Class for extracting features from a HDF5 file
    """
    def __init__(self):
        pass

    @staticmethod
    def extract_attribute(file: h5py.File,group_path: str, attribute: str) -> str:
        """Extract an attribute from a HDF5 file

        Args:
            file (h5py.File): An opened HDF5 file
            group_path (str): Path to the HDF5 Group containing the attribute
            attribute (str): Attribute to look for

        Raises:
            ValueError: _description_

        Returns:
            str: Value of the attribute stored in the file
        """
        try:
            group = file[group_path]

            if not isinstance(group,h5py.Group):
                raise ValueError

            return group.attrs[attribute].strip()

        except Exception:
            return None

    def extract_min_max_coords(file: h5py.File,dataset_path: str) -> Tuple[np.ndarray,np.ndarray]:
        """Read a set of 3d coordinates from the HDF5 file and keep only the minimum and maximum coordinates

        Args:
            file (h5py.File): HDF5 file containing coordinates
            dataset_path (str): Datasets containing the 3D coordinates

        Raises:
            ValueError: _description_

        Returns:
            Tuple[np.ndarray,np.ndarray]: Minimum coordinate and Maximum coordinate
        """
        dataset: h5py.Dataset = file[dataset_path]
        shape = dataset.shape

        # Ensure dataset is at least 2D and has 3 columns
        if len(shape) < 2 or shape[1] != 3:
            raise ValueError(f"Expected a dataset with shape (N,3), but got {shape}")

        #initialise
        CHUNK_SIZE = 1000 #TODO where to put chunk_size???
        min_coord = np.full(3, np.inf)
        max_coord = np.full(3, -np.inf)
        
        for chunk_start in range(0,shape[0],CHUNK_SIZE):
            chunk_end = min(chunk_start + CHUNK_SIZE, shape[0])
            data_chunk = dataset[chunk_start:chunk_end, :]

            min_coord_in_chunk: np.ndarray = np.amin(data_chunk,axis=0)
            max_coord_in_chunk: np.ndarray = np.amax(data_chunk,axis=0)

            #in-place as memory allocations are expensive in Python
            np.minimum(min_coord,min_coord_in_chunk,out=min_coord)
            np.maximum(max_coord,max_coord_in_chunk,out=max_coord)

        return min_coord, max_coord

#TODO Should NekAutoExtractor really be in this package???
class NekAutoExtractor:
    """Nektar auto extractor of fields in Nektar dataset files
    """
    def __init__(self,session_file: str,geometry_file: str,output_file: List[str]):
        """Class initialiser

        Args:
            session_file (str): Session file path
            geometry_file (str): Geometry file path
            output_file (List[str]): List of output file paths
        """
        self.session_file = session_file
        self.geometry_file = geometry_file
        self.output_file = output_file

    def extract_data(self) -> Dict[str,str]:
        """Extract data from the files

        Returns:
            Dict[str,str]: Data extracted from the Nektar++ datasets
        """
        results = {}

        with h5py.File(self.output_file) as f:
            if version := HDF5Extractor.extract_attribute(f,"NEKTAR/Metadata/Provenance","NektarVersion"):
                results["VERSION"] = version 
            
            if git_hash := HDF5Extractor.extract_attribute(f,"NEKTAR/Metadata/Provenance","GitSHA1"):
                results["GITHASH"] = git_hash
        
        with h5py.File(self.geometry_file) as f:
            min_coords1,max_coords1 = HDF5Extractor.extract_min_max_coords(f,"NEKTAR/GEOMETRY/MESH/VERT")
            min_coords2,max_coords2 = HDF5Extractor.extract_min_max_coords(f,"NEKTAR/GEOMETRY/MESH/CURVE_NODES")#mot needed

            min_coords = np.minimum(min_coords1,min_coords2)
            max_coords = np.maximum(max_coords1,max_coords2)

            results["MAX_COORD"] = max_coords
            results["MIN_COORD"] = min_coords

        return results

if __name__ == "__main__":
    with h5py.File("tests/datasets/ADRSolver/ADR_2D_TriQuad.fld") as f:
        v = HDF5Extractor.extract_attribute(f,"NEKTAR/Metadata/Provenance","NektarVersion")
        print(v)

    with h5py.File("tests/datasets/ADRSolver/ADR_2D_TriQuad.nekg") as f:
        coords = HDF5Extractor.extract_min_max_coords(f,"NEKTAR/GEOMETRY/MESH/VERT")
        print(coords)